@@ -1,13 +1,11 @@
-# -*- coding: utf-8 -*-
-
 from collections import namedtuple
 
-__version__ = '0.27.0'
+__version__ = '0.36.0'
 
 App = namedtuple('App', ['instance', 'base_url', 'client_id', 'client_secret'])
 User = namedtuple('User', ['instance', 'username', 'access_token'])
 
-DEFAULT_INSTANCE = 'mastodon.social'
+DEFAULT_INSTANCE = 'https://mastodon.social'
 
 CLIENT_NAME = 'toot - a Mastodon CLI client'
 CLIENT_WEBSITE = 'https://github.com/ihabunek/toot'
