@@ -1,25 +1,75 @@
-# -*- coding: utf-8 -*-
-
 import logging
 import os
+import re
 import shutil
 import sys
 
-from argparse import ArgumentParser, FileType, ArgumentTypeError
+from argparse import ArgumentParser, FileType, ArgumentTypeError, Action
 from collections import namedtuple
+from itertools import chain
 from toot import config, commands, CLIENT_NAME, CLIENT_WEBSITE, __version__
 from toot.exceptions import ApiError, ConsoleError
 from toot.output import print_out, print_err
 
-VISIBILITY_CHOICES = ['public', 'unlisted', 'private', 'direct']
+VISIBILITY_CHOICES = ["public", "unlisted", "private", "direct"]
+VISIBILITY_CHOICES_STR = ", ".join(f"'{v}'" for v in VISIBILITY_CHOICES)
+
+PRIVACY_CHOICES = ["public", "unlisted", "private"]
+PRIVACY_CHOICES_STR = ", ".join(f"'{v}'" for v in PRIVACY_CHOICES)
+
+
+class BooleanOptionalAction(Action):
+    """
+    Backported from argparse. This action is available since Python 3.9.
+    https://github.com/python/cpython/blob/3.11/Lib/argparse.py
+    """
+    def __init__(self,
+                 option_strings,
+                 dest,
+                 default=None,
+                 type=None,
+                 choices=None,
+                 required=False,
+                 help=None,
+                 metavar=None):
+
+        _option_strings = []
+        for option_string in option_strings:
+            _option_strings.append(option_string)
+
+            if option_string.startswith('--'):
+                option_string = '--no-' + option_string[2:]
+                _option_strings.append(option_string)
+
+        super().__init__(
+            option_strings=_option_strings,
+            dest=dest,
+            nargs=0,
+            default=default,
+            type=type,
+            choices=choices,
+            required=required,
+            help=help,
+            metavar=metavar)
+
+    def __call__(self, parser, namespace, values, option_string=None):
+        if option_string in self.option_strings:
+            setattr(namespace, self.dest, not option_string.startswith('--no-'))
+
+    def format_usage(self):
+        return ' | '.join(self.option_strings)
+
+
+def get_default_visibility():
+    return os.getenv("TOOT_POST_VISIBILITY", "public")
 
 
 def language(value):
     """Validates the language parameter"""
-    if len(value) != 3:
+    if len(value) != 2:
         raise ArgumentTypeError(
-            "Invalid language specified: '{}'. Expected a 3 letter "
-            "abbreviation according to ISO 639-2 standard.".format(value)
+            "Invalid language. Expected a 2 letter abbreviation according to "
+            "the ISO 639-1 standard."
         )
 
     return value
@@ -33,6 +83,14 @@ def visibility(value):
     return value
 
 
+def privacy(value):
+    """Validates the privacy parameter"""
+    if value not in PRIVACY_CHOICES:
+        raise ValueError(f"Invalid privacy value. Expected one of {PRIVACY_CHOICES_STR}.")
+
+    return value
+
+
 def timeline_count(value):
     n = int(value)
     if not 0 < n <= 20:
@@ -40,6 +98,46 @@ def timeline_count(value):
     return n
 
 
+DURATION_UNITS = {
+    "m": 60,
+    "h": 60 * 60,
+    "d": 60 * 60 * 24,
+}
+
+
+DURATION_EXAMPLES = """e.g. "1 day", "2 hours 30 minutes", "5 minutes 30
+seconds" or any combination of above. Shorthand: "1d", "2h30m", "5m30s\""""
+
+
+def duration(value: str):
+    match = re.match(r"""^
+        (([0-9]+)\s*(days|day|d))?\s*
+        (([0-9]+)\s*(hours|hour|h))?\s*
+        (([0-9]+)\s*(minutes|minute|m))?\s*
+        (([0-9]+)\s*(seconds|second|s))?\s*
+    $""", value, re.X)
+
+    if not match:
+        raise ArgumentTypeError(f"Invalid duration: {value}")
+
+    days = match.group(2)
+    hours = match.group(5)
+    minutes = match.group(8)
+    seconds = match.group(11)
+
+    days = int(match.group(2) or 0) * 60 * 60 * 24
+    hours = int(match.group(5) or 0) * 60 * 60
+    minutes = int(match.group(8) or 0) * 60
+    seconds = int(match.group(11) or 0)
+
+    duration = days + hours + minutes + seconds
+
+    if duration == 0:
+        raise ArgumentTypeError("Empty duration")
+
+    return duration
+
+
 def editor(value):
     if not value:
         raise ArgumentTypeError(
@@ -75,6 +173,11 @@ common_args = [
         "action": 'store_true',
         "default": False,
     }),
+    (["--verbose"], {
+        "help": "show extra detail in debug log; used with --debug",
+        "action": 'store_true',
+        "default": False,
+    }),
 ]
 
 # Arguments added to commands which require authentication
@@ -111,6 +214,19 @@ status_id_arg = (["status_id"], {
     "type": str,
 })
 
+visibility_arg = (["-v", "--visibility"], {
+    "type": visibility,
+    "default": get_default_visibility(),
+    "help": f"Post visibility. One of: {VISIBILITY_CHOICES_STR}. Defaults to "
+            f"'{get_default_visibility()}' which can be overridden by setting "
+            "the TOOT_POST_VISIBILITY environment variable",
+})
+
+tag_arg = (["tag_name"], {
+    "type": str,
+    "help": "tag name, e.g. Caturday, or \"#Caturday\"",
+})
+
 # Arguments for selecting a timeline (see `toot.commands.get_timeline_generator`)
 common_timeline_args = [
     (["-p", "--public"], {
@@ -137,7 +253,7 @@ common_timeline_args = [
     }),
 ]
 
-timeline_args = common_timeline_args + [
+timeline_and_bookmark_args = [
     (["-c", "--count"], {
         "type": timeline_count,
         "help": "number of toots to show per page (1-20, default 10).",
@@ -155,6 +271,8 @@ timeline_args = common_timeline_args + [
     }),
 ]
 
+timeline_args = common_timeline_args + timeline_and_bookmark_args
+
 AUTH_COMMANDS = [
     Command(
         name="login",
@@ -186,13 +304,72 @@ AUTH_COMMANDS = [
         arguments=[],
         require_auth=False,
     ),
+    Command(
+        name="env",
+        description="Print environment information for inclusion in bug reports.",
+        arguments=[],
+        require_auth=False,
+    ),
+    Command(
+        name="update_account",
+        description="Update your account details",
+        arguments=[
+            (["--display-name"], {
+                "type": str,
+                "help": "The display name to use for the profile.",
+            }),
+            (["--note"], {
+                "type": str,
+                "help": "The account bio.",
+            }),
+            (["--avatar"], {
+                "type": FileType("rb"),
+                "help": "Path to the avatar image to set.",
+            }),
+            (["--header"], {
+                "type": FileType("rb"),
+                "help": "Path to the header image to set.",
+            }),
+            (["--bot"], {
+                "action": BooleanOptionalAction,
+                "help": "Whether the account has a bot flag.",
+            }),
+            (["--discoverable"], {
+                "action": BooleanOptionalAction,
+                "help": "Whether the account should be shown in the profile directory.",
+            }),
+            (["--locked"], {
+                "action": BooleanOptionalAction,
+                "help": "Whether manual approval of follow requests is required.",
+            }),
+            (["--privacy"], {
+                "type": privacy,
+                "help": f"Default post privacy for authored statuses. One of: {PRIVACY_CHOICES_STR}."
+            }),
+            (["--sensitive"], {
+                "action": BooleanOptionalAction,
+                "help": "Whether to mark authored statuses as sensitive by default."
+            }),
+            (["--language"], {
+                "type": language,
+                "help": "Default language to use for authored statuses (ISO 6391)."
+            }),
+        ],
+        require_auth=True,
+    ),
 ]
 
 TUI_COMMANDS = [
     Command(
         name="tui",
         description="Launches the toot terminal user interface",
-        arguments=[],
+        arguments=[
+            (["--relative-datetimes"], {
+                "action": "store_true",
+                "default": False,
+                "help": "Show relative datetimes in status list.",
+            }),
+        ],
         require_auth=True,
     ),
 ]
@@ -280,6 +457,12 @@ READ_COMMANDS = [
         arguments=timeline_args,
         require_auth=True,
     ),
+    Command(
+        name="bookmarks",
+        description="Show bookmarked posts",
+        arguments=timeline_and_bookmark_args,
+        require_auth=True,
+    ),
 ]
 
 POST_COMMANDS = [
@@ -303,11 +486,13 @@ POST_COMMANDS = [
                 "help": "plain-text description of the media for accessibility "
                         "purposes, one per attached media"
             }),
-            (["-v", "--visibility"], {
-                "type": visibility,
-                "default": "public",
-                "help": 'post visibility, one of: %s' % ", ".join(VISIBILITY_CHOICES),
+            (["--thumbnail"], {
+                "action": "append",
+                "type": FileType("rb"),
+                "help": "path to an image file to serve as media thumbnail, "
+                        "one per attached media"
             }),
+            visibility_arg,
             (["-s", "--sensitive"], {
                 "action": 'store_true',
                 "default": False,
@@ -337,10 +522,37 @@ POST_COMMANDS = [
                 "help": "ISO 8601 Datetime at which to schedule a status. Must "
                         "be at least 5 minutes in the future.",
             }),
+            (["--scheduled-in"], {
+                "type": duration,
+                "help": f"""Schedule the toot to be posted after a given amount
+                        of time, {DURATION_EXAMPLES}. Must be at least 5
+                        minutes.""",
+            }),
             (["-t", "--content-type"], {
                 "type": str,
                 "help": "MIME type for the status text (not supported on all instances)",
             }),
+            (["--poll-option"], {
+                "action": "append",
+                "type": str,
+                "help": "Possible answer to the poll"
+            }),
+            (["--poll-expires-in"], {
+                "type": duration,
+                "help": f"""Duration that the poll should be open,
+                        {DURATION_EXAMPLES}. Defaults to 24h.""",
+                "default": 24 * 60 * 60,
+            }),
+            (["--poll-multiple"], {
+                "action": "store_true",
+                "default": False,
+                "help": "Allow multiple answers to be selected."
+            }),
+            (["--poll-hide-totals"], {
+                "action": "store_true",
+                "default": False,
+                "help": "Hide vote counts until the poll ends. Defaults to false."
+            }),
         ],
         require_auth=True,
     ),
@@ -383,7 +595,7 @@ STATUS_COMMANDS = [
     Command(
         name="reblog",
         description="Reblog a status",
-        arguments=[status_id_arg],
+        arguments=[status_id_arg, visibility_arg],
         require_auth=True,
     ),
     Command(
@@ -410,6 +622,18 @@ STATUS_COMMANDS = [
         arguments=[status_id_arg],
         require_auth=True,
     ),
+    Command(
+        name="bookmark",
+        description="Bookmark a status",
+        arguments=[status_id_arg],
+        require_auth=True,
+    ),
+    Command(
+        name="unbookmark",
+        description="Unbookmark a status",
+        arguments=[status_id_arg],
+        require_auth=True,
+    ),
 ]
 
 ACCOUNTS_COMMANDS = [
@@ -429,6 +653,22 @@ ACCOUNTS_COMMANDS = [
         ],
         require_auth=True,
     ),
+    Command(
+        name="following",
+        description="List accounts followed by the given account",
+        arguments=[
+            account_arg,
+        ],
+        require_auth=True,
+    ),
+    Command(
+        name="followers",
+        description="List accounts following the given account",
+        arguments=[
+            account_arg,
+        ],
+        require_auth=True,
+    ),
     Command(
         name="mute",
         description="Mute an account",
@@ -463,25 +703,47 @@ ACCOUNTS_COMMANDS = [
     ),
 ]
 
-COMMANDS = AUTH_COMMANDS + READ_COMMANDS + TUI_COMMANDS + POST_COMMANDS + STATUS_COMMANDS + ACCOUNTS_COMMANDS
+TAG_COMMANDS = [
+    Command(
+        name="tags_followed",
+        description="List hashtags you follow",
+        arguments=[],
+        require_auth=True,
+    ),
+    Command(
+        name="tags_follow",
+        description="Follow a hashtag",
+        arguments=[tag_arg],
+        require_auth=True,
+    ),
+    Command(
+        name="tags_unfollow",
+        description="Unfollow a hashtag",
+        arguments=[tag_arg],
+        require_auth=True,
+    ),
+]
+
+COMMAND_GROUPS = [
+    ("Authentication", AUTH_COMMANDS),
+    ("TUI", TUI_COMMANDS),
+    ("Read", READ_COMMANDS),
+    ("Post", POST_COMMANDS),
+    ("Status", STATUS_COMMANDS),
+    ("Accounts", ACCOUNTS_COMMANDS),
+    ("Hashtags", TAG_COMMANDS),
+]
+
+COMMANDS = list(chain(*[commands for _, commands in COMMAND_GROUPS]))
 
 
 def print_usage():
-    max_name_len = max(len(command.name) for command in COMMANDS)
-
-    groups = [
-        ("Authentication", AUTH_COMMANDS),
-        ("TUI", TUI_COMMANDS),
-        ("Read", READ_COMMANDS),
-        ("Post", POST_COMMANDS),
-        ("Status", STATUS_COMMANDS),
-        ("Accounts", ACCOUNTS_COMMANDS),
-    ]
+    max_name_len = max(len(name) for name, _ in COMMAND_GROUPS)
 
     print_out("<green>{}</green>".format(CLIENT_NAME))
     print_out("<blue>v{}</blue>".format(__version__))
 
-    for name, cmds in groups:
+    for name, cmds in COMMAND_GROUPS:
         print_out("")
         print_out(name + ":")
 
@@ -491,7 +753,7 @@ def print_usage():
 
     print_out("")
     print_out("To get help for each command run:")
-    print_out("  <yellow>toot <command> --help</yellow>")
+    print_out("  <yellow>toot \\<command> --help</yellow>")
     print_out("")
     print_out("<green>{}</green>".format(CLIENT_WEBSITE))
 
@@ -516,8 +778,8 @@ def run_command(app, user, name, args):
     command = next((c for c in COMMANDS if c.name == name), None)
 
     if not command:
-        print_err("Unknown command '{}'\n".format(name))
-        print_usage()
+        print_err(f"Unknown command '{name}'")
+        print_out("Run <yellow>toot --help</yellow> to show a list of available commands.")
         return
 
     parser = get_argument_parser(name, command)
@@ -547,11 +809,12 @@ def main():
     if "--debug" in sys.argv:
         filename = os.getenv("TOOT_LOG_FILE")
         logging.basicConfig(level=logging.DEBUG, filename=filename)
+        logging.getLogger("urllib3").setLevel(logging.INFO)
 
     command_name = sys.argv[1] if len(sys.argv) > 1 else None
     args = sys.argv[2:]
 
-    if not command_name:
+    if not command_name or command_name == "--help":
         return print_usage()
 
     user, app = config.get_active_user_app()
