@@ -4,14 +4,17 @@ import urwid
 from concurrent.futures import ThreadPoolExecutor
 
 from toot import api, config, __version__
+from toot.console import get_default_visibility
+from toot.exceptions import ApiError
 
 from .compose import StatusComposer
 from .constants import PALETTE
 from .entities import Status
-from .overlays import ExceptionStackTrace, GotoMenu, Help, StatusSource, StatusLinks
-from .overlays import StatusDeleteConfirmation
+from .overlays import ExceptionStackTrace, GotoMenu, Help, StatusSource, StatusLinks, StatusZoom
+from .overlays import StatusDeleteConfirmation, Account
+from .poll import Poll
 from .timeline import Timeline
-from .utils import parse_content_links, show_media
+from .utils import parse_content_links, show_media, copy_to_clipboard
 
 logger = logging.getLogger(__name__)
 
@@ -71,10 +74,10 @@ class TUI(urwid.Frame):
     """Main TUI frame."""
 
     @classmethod
-    def create(cls, app, user):
+    def create(cls, app, user, args):
         """Factory method, sets up TUI and an event loop."""
 
-        tui = cls(app, user)
+        tui = cls(app, user, args)
         loop = urwid.MainLoop(
             tui,
             palette=PALETTE,
@@ -85,9 +88,10 @@ class TUI(urwid.Frame):
 
         return tui
 
-    def __init__(self, app, user):
+    def __init__(self, app, user, args):
         self.app = app
         self.user = user
+        self.args = args
         self.config = config.load_config()
 
         self.loop = None  # set in `create`
@@ -106,11 +110,14 @@ class TUI(urwid.Frame):
         self.timeline = None
         self.overlay = None
         self.exception = None
+        self.can_translate = False
+        self.screen = urwid.raw_display.Screen()
 
         super().__init__(self.body, header=self.header, footer=self.footer)
 
     def run(self):
         self.loop.set_alarm_in(0, lambda *args: self.async_load_instance())
+        self.loop.set_alarm_in(0, lambda *args: self.async_load_followed_tags())
         self.loop.set_alarm_in(0, lambda *args: self.async_load_timeline(
             is_initial=True, timeline_name="home"))
         self.loop.run()
@@ -150,7 +157,7 @@ class TUI(urwid.Frame):
 
         def _default_error_callback(ex):
             self.exception = ex
-            self.footer.set_error_message("An exeption occured, press E to view")
+            self.footer.set_error_message("An exception occurred, press X to view")
 
         _error_callback = error_callback or _default_error_callback
 
@@ -170,6 +177,9 @@ class TUI(urwid.Frame):
         return future
 
     def connect_default_timeline_signals(self, timeline):
+        def _account(timeline, account_id):
+            self.show_account(account_id)
+
         def _compose(*args):
             self.show_compose()
 
@@ -192,16 +202,35 @@ class TUI(urwid.Frame):
         def _menu(timeline, status):
             self.show_context_menu(status)
 
+        def _poll(timeline, status):
+            self.show_poll(status)
+
+        def _zoom(timeline, status_details):
+            self.show_status_zoom(status_details)
+
+        def _clear(*args):
+            self.clear_screen()
+
+        def _copy(timeline, status):
+            self.copy_status(status)
+
+        urwid.connect_signal(timeline, "account", _account)
+        urwid.connect_signal(timeline, "bookmark", self.async_toggle_bookmark)
         urwid.connect_signal(timeline, "compose", _compose)
         urwid.connect_signal(timeline, "delete", _delete)
         urwid.connect_signal(timeline, "favourite", self.async_toggle_favourite)
         urwid.connect_signal(timeline, "focus", self.refresh_footer)
         urwid.connect_signal(timeline, "media", _media)
         urwid.connect_signal(timeline, "menu", _menu)
+        urwid.connect_signal(timeline, "poll", _poll)
         urwid.connect_signal(timeline, "reblog", self.async_toggle_reblog)
         urwid.connect_signal(timeline, "reply", _reply)
         urwid.connect_signal(timeline, "source", _source)
         urwid.connect_signal(timeline, "links", _links)
+        urwid.connect_signal(timeline, "zoom", _zoom)
+        urwid.connect_signal(timeline, "translate", self.async_translate)
+        urwid.connect_signal(timeline, "clear-screen", _clear)
+        urwid.connect_signal(timeline, "copy-status", _copy)
 
     def build_timeline(self, name, statuses, local):
         def _close(*args):
@@ -228,7 +257,7 @@ class TUI(urwid.Frame):
             self.loop.set_alarm_in(5, lambda *args: self.footer.clear_message())
             config.save_config(self.config)
 
-        timeline = Timeline(name, statuses)
+        timeline = Timeline(name, statuses, self.can_translate, self.followed_tags)
 
         self.connect_default_timeline_signals(timeline)
         urwid.connect_signal(timeline, "next", _next)
@@ -257,7 +286,8 @@ class TUI(urwid.Frame):
         statuses = ancestors + [status] + descendants
         focus = len(ancestors)
 
-        timeline = Timeline("thread", statuses, focus, is_thread=True)
+        timeline = Timeline("thread", statuses, self.can_translate,
+                            self.followed_tags, focus, is_thread=True)
 
         self.connect_default_timeline_signals(timeline)
         urwid.connect_signal(timeline, "close", _close)
@@ -299,16 +329,49 @@ class TUI(urwid.Frame):
         Attempt to update max_toot_chars from instance data.
         Does not work on vanilla Mastodon, works on Pleroma.
         See: https://github.com/tootsuite/mastodon/issues/4915
+
+        Also attempt to update translation flag from instance
+        data. Translation is only present on Mastodon 4+ servers
+        where the administrator has enabled this feature.
+        See: https://github.com/mastodon/mastodon/issues/19328
         """
         def _load_instance():
-            return api.get_instance(self.app.instance)
+            return api.get_instance(self.app.base_url)
 
         def _done(instance):
             if "max_toot_chars" in instance:
                 self.max_toot_chars = instance["max_toot_chars"]
 
+            if "translation" in instance:
+                # instance is advertising translation service
+                self.can_translate = instance["translation"]["enabled"]
+            elif "version" in instance:
+                # fallback check:
+                # get the major version number of the server
+                # this works for Mastodon and Pleroma version strings
+                # Mastodon versions < 4 do not have translation service
+                # Revisit this logic if Pleroma implements translation
+                ch = instance["version"][0]
+                self.can_translate = int(ch) > 3 if ch.isnumeric() else False
+
         return self.run_in_thread(_load_instance, done_callback=_done)
 
+    def async_load_followed_tags(self):
+        def _load_tag_list():
+            try:
+                return api.followed_tags(self.app, self.user)
+            except ApiError:
+                # not supported by all Mastodon servers so fail silently if necessary
+                return []
+
+        def _done_tag_list(tags):
+            if len(tags) > 0:
+                self.followed_tags = [t["name"] for t in tags]
+            else:
+                self.followed_tags = []
+
+        self.run_in_thread(_load_tag_list, done_callback=_done_tag_list)
+
     def refresh_footer(self, timeline):
         """Show status details in footer."""
         status, index, count = timeline.get_focused_status_with_counts()
@@ -323,15 +386,38 @@ class TUI(urwid.Frame):
             title="Status source",
         )
 
+    def clear_screen(self):
+        self.loop.screen.clear()
+
     def show_links(self, status):
-        links = parse_content_links(status.data["content"]) if status else []
+        links = parse_content_links(status.original.data["content"]) if status else []
+        post_attachments = status.original.data["media_attachments"] or []
+        reblog_attachments = (status.data["reblog"]["media_attachments"] if status.data["reblog"] else None) or []
+
+        for a in post_attachments + reblog_attachments:
+            url = a["remote_url"] or a["url"]
+            links.append((url, a["description"] if a["description"] else url))
+
+        def _clear(*args):
+            self.clear_screen()
+
         if links:
+            links = list(set(links))  # deduplicate links
+            links = sorted(links, key=lambda link: link[0])  # sort alphabetically by URL
+            sl_widget = StatusLinks(links)
+            urwid.connect_signal(sl_widget, "clear-screen", _clear)
             self.open_overlay(
-                widget=StatusLinks(links),
+                widget=sl_widget,
                 title="Status links",
                 options={"height": len(links) + 2},
             )
 
+    def show_status_zoom(self, status_details):
+        self.open_overlay(
+            widget=StatusZoom(status_details),
+            title="Status zoom",
+        )
+
     def show_exception(self, exception):
         self.open_overlay(
             widget=ExceptionStackTrace(exception),
@@ -345,7 +431,7 @@ class TUI(urwid.Frame):
         def _post(timeline, *args):
             self.post_status(*args)
 
-        composer = StatusComposer(self.max_toot_chars, in_reply_to)
+        composer = StatusComposer(self.max_toot_chars, self.user.username, in_reply_to)
         urwid.connect_signal(composer, "close", _close)
         urwid.connect_signal(composer, "post", _post)
         self.open_overlay(composer, title="Compose status")
@@ -357,17 +443,29 @@ class TUI(urwid.Frame):
             lambda x: self.goto_home_timeline())
         urwid.connect_signal(menu, "public_timeline",
             lambda x, local: self.goto_public_timeline(local))
+        urwid.connect_signal(menu, "bookmark_timeline",
+            lambda x, local: self.goto_bookmarks())
+        urwid.connect_signal(menu, "notification_timeline",
+            lambda x, local: self.goto_notifications())
+        urwid.connect_signal(menu, "conversation_timeline",
+            lambda x, local: self.goto_conversations())
         urwid.connect_signal(menu, "hashtag_timeline",
             lambda x, tag, local: self.goto_tag_timeline(tag, local=local))
 
         self.open_overlay(menu, title="Go to", options=dict(
             align="center", width=("relative", 60),
-            valign="middle", height=9 + len(user_timelines),
+            valign="middle", height=11 + len(user_timelines),
         ))
 
     def show_help(self):
         self.open_overlay(Help(), title="Help")
 
+    def show_poll(self, status):
+        self.open_overlay(
+            widget=Poll(self.app, self.user, status),
+            title="Poll",
+        )
+
     def goto_home_timeline(self):
         self.timeline_generator = api.home_timeline_generator(
             self.app, self.user, limit=40)
@@ -380,6 +478,27 @@ class TUI(urwid.Frame):
         promise = self.async_load_timeline(is_initial=True, timeline_name="public")
         promise.add_done_callback(lambda *args: self.close_overlay())
 
+    def goto_bookmarks(self):
+        self.timeline_generator = api.bookmark_timeline_generator(
+            self.app, self.user, limit=40)
+        promise = self.async_load_timeline(is_initial=True, timeline_name="bookmarks")
+        promise.add_done_callback(lambda *args: self.close_overlay())
+
+    def goto_notifications(self):
+        self.timeline_generator = api.notification_timeline_generator(
+            self.app, self.user, limit=40)
+        promise = self.async_load_timeline(is_initial=True, timeline_name="notifications")
+        promise.add_done_callback(lambda *args: self.close_overlay())
+
+    def goto_conversations(self):
+        self.timeline_generator = api.conversation_timeline_generator(
+            self.app, self.user, limit=40
+        )
+        promise = self.async_load_timeline(
+            is_initial=True, timeline_name="conversations"
+        )
+        promise.add_done_callback(lambda *args: self.close_overlay())
+
     def goto_tag_timeline(self, tag, local):
         self.timeline_generator = api.tag_timeline_generator(
             self.app, self.user, tag, local=local, limit=40)
@@ -420,13 +539,19 @@ class TUI(urwid.Frame):
             in_reply_to_id=in_reply_to_id)
         status = self.make_status(data)
 
-        # TODO: instead of this, fetch new items from the timeline?
-        self.timeline.prepend_status(status)
-        self.timeline.focus_status(status)
+        # TODO: fetch new items from the timeline?
 
         self.footer.set_message("Status posted {} \\o/".format(status.id))
         self.close_overlay()
 
+    def show_account(self, account_id):
+        account = api.whois(self.app, self.user, account_id)
+        relationship = api.get_relationship(self.app, self.user, account_id)
+        self.open_overlay(
+            widget=Account(self.app, self.user, account, relationship),
+            title="Account",
+        )
+
     def async_toggle_favourite(self, timeline, status):
         def _favourite():
             logger.info("Favouriting {}".format(status))
@@ -451,7 +576,7 @@ class TUI(urwid.Frame):
     def async_toggle_reblog(self, timeline, status):
         def _reblog():
             logger.info("Reblogging {}".format(status))
-            api.reblog(self.app, self.user, status.id)
+            api.reblog(self.app, self.user, status.id, visibility=get_default_visibility())
 
         def _unreblog():
             logger.info("Unreblogging {}".format(status))
@@ -464,11 +589,72 @@ class TUI(urwid.Frame):
             new_status = self.make_status(new_data)
             timeline.update_status(new_status)
 
+        # Check if status is rebloggable
+        no_reblog_because_private = status.visibility == "private" and not status.is_mine
+        no_reblog_because_direct = status.visibility == "direct"
+        if no_reblog_because_private or no_reblog_because_direct:
+            self.footer.set_error_message("You may not reblog this {} status".format(status.visibility))
+            return
+
         self.run_in_thread(
             _unreblog if status.reblogged else _reblog,
             done_callback=_done
         )
 
+    def async_translate(self, timeline, status):
+        def _translate():
+            logger.info("Translating {}".format(status))
+            self.footer.set_message("Translating status {}".format(status.original.id))
+
+            try:
+                response = api.translate(self.app, self.user, status.original.id)
+                if response["content"]:
+                    self.footer.set_message("Status translated")
+                else:
+                    self.footer.set_error_message("Server returned empty translation")
+                    response = None
+            except Exception:
+                response = None
+                self.footer.set_error_message("Translate server error")
+
+            self.loop.set_alarm_in(3, lambda *args: self.footer.clear_message())
+            return response
+
+        def _done(response):
+            if response is not None:
+                status.original.translation = response["content"]
+                status.original.translated_from = response["detected_source_language"]
+                status.original.show_translation = True
+                timeline.update_status(status)
+
+        # If already translated, toggle showing translation
+        if status.original.translation:
+            status.original.show_translation = not status.original.show_translation
+            timeline.update_status(status)
+        else:
+            self.run_in_thread(_translate, done_callback=_done)
+
+    def async_toggle_bookmark(self, timeline, status):
+        def _bookmark():
+            logger.info("Bookmarking {}".format(status))
+            api.bookmark(self.app, self.user, status.id)
+
+        def _unbookmark():
+            logger.info("Unbookmarking {}".format(status))
+            api.unbookmark(self.app, self.user, status.id)
+
+        def _done(loop):
+            # Create a new Status with flipped bookmarked flag
+            new_data = status.data
+            new_data["bookmarked"] = not status.bookmarked
+            new_status = self.make_status(new_data)
+            timeline.update_status(new_status)
+
+        self.run_in_thread(
+            _unbookmark if status.bookmarked else _bookmark,
+            done_callback=_done
+        )
+
     def async_delete_status(self, timeline, status):
         def _delete():
             api.delete_status(self.app, self.user, status.id)
@@ -478,6 +664,12 @@ class TUI(urwid.Frame):
 
         return self.run_in_thread(_delete, done_callback=_done)
 
+    def copy_status(self, status):
+        # TODO: copy a better version of status content
+        # including URLs
+        copy_to_clipboard(self.screen, status.original.data["content"])
+        self.footer.set_message(f"Status {status.original.id} copied")
+
     # --- Overlay handling -----------------------------------------------------
 
     default_overlay_options = dict(
@@ -502,12 +694,14 @@ class TUI(urwid.Frame):
     def close_overlay(self):
         self.body = self.overlay.bottom_w
         self.overlay = None
+        if self.timeline:
+            self.timeline.refresh_status_details()
 
     # --- Keys -----------------------------------------------------------------
 
     def unhandled_input(self, key):
         # TODO: this should not be in unhandled input
-        if key in ('e', 'E'):
+        if key in ('x', 'X'):
             if self.exception:
                 self.show_exception(self.exception)
 
@@ -519,6 +713,12 @@ class TUI(urwid.Frame):
             if not self.overlay:
                 self.show_help()
 
+        elif key == ',':
+            if not self.overlay:
+                self.timeline_generator = api.home_timeline_generator(
+                    self.app, self.user, limit=40)
+                self.async_load_timeline(is_initial=True, timeline_name=self.timeline.name)
+
         elif key == 'esc':
             if self.overlay:
                 self.close_overlay()
