@@ -19,6 +19,7 @@ PALETTE = [
 
     # Functional
     ('hashtag', 'light cyan,bold', ''),
+    ('followed_hashtag', 'yellow,bold', ''),
     ('link', ',italics', ''),
     ('link_focused', ',italics', 'dark magenta'),
 
@@ -34,7 +35,9 @@ PALETTE = [
     ('green_selected', 'white,bold', 'dark green'),
     ('yellow', 'yellow', ''),
     ('yellow_bold', 'yellow,bold', ''),
+    ('red', 'dark red', ''),
     ('warning', 'light red', ''),
+    ('white_bold', 'white,bold', '')
 ]
 
 VISIBILITY_OPTIONS = [
